# -*- coding: utf-8 -*-

""" Scrapy downloader middleware """

import logging

from scrapy.exceptions import NotConfigured
from twisted.internet import reactor, defer

from .utils import identity

LOGGER = logging.getLogger(__name__)


class DelayedRetry:
    """ retry requests with a delay """

    @classmethod
    def from_crawler(cls, crawler):
        """ init from crawler settings """

        settings = crawler.settings

        http_codes = frozenset(
            int(x) for x in settings.getlist("DELAYED_RETRY_HTTP_CODES")
        )
        delay = settings.getfloat("DELAYED_RETRY_DELAY", 0.1)
        times = settings.getint("DELAYED_RETRY_TIMES", -1)
        backoff = settings.getbool("DELAYED_RETRY_BACKOFF")
        backoff_max_delay = settings.getfloat(
            "DELAYED_RETRY_BACKOFF_MAX_DELAY", 10 * delay
        )
        priority_adjust = settings.getint(
            "DELAYED_RETRY_PRIORITY_ADJUST", settings.getint("RETRY_PRIORITY_ADJUST")
        )

        if not settings.getbool("DELAYED_RETRY_ENABLED") or not http_codes:
            raise NotConfigured

        return cls(
            http_codes, delay, times, backoff, backoff_max_delay, priority_adjust
        )

    def __init__(
        self,
        http_codes,
        delay,
        times=-1,
        backoff=False,
        backoff_max_delay=None,
        priority_adjust=0,
    ):
        self.http_codes = http_codes
        self.delay = delay
        self.times = times
        self.backoff = backoff
        self.backoff_max_delay = backoff_max_delay
        self.priority_adjust = priority_adjust

    # pylint: disable=unused-argument
    def process_response(self, request, response, spider):
        """ retry certain requests with delay """

        if request.meta.get("dont_retry", False):
            return response

        if response.status in self.http_codes:
            return self._retry(request) or response

        return response

    def _retry(self, request):
        retries = request.meta.get("retry_times", 0)
        delay = request.meta.get("retry_delay", self.delay)

        max_retry_times = request.meta.get("max_retry_times", self.times)

        if retries >= max_retry_times >= 0:
            LOGGER.info(
                "request has already been retried %d times, but is allowed only %d retries",
                retries,
                max_retry_times,
            )
            return None

        req = request.copy()
        req.meta["retry_times"] = retries + 1
        req.meta["retry_delay"] = (
            min(2 * delay, self.backoff_max_delay) if self.backoff else delay
        )
        req.dont_filter = True
        req.priority = request.priority + self.priority_adjust

        LOGGER.debug("retry request %r in %d seconds", req, delay)

        deferred = defer.Deferred()
        deferred.addCallback(identity)
        # pylint: disable=no-member
        reactor.callLater(delay, deferred.callback, req)

        return deferred



# -*- coding: utf-8 -*-

""" Scrapy item loaders """

import jmespath

from pytility import normalize_space
from scrapy.loader import ItemLoader
from scrapy.loader.processors import TakeFirst, MapCompose
from scrapy.utils.misc import arg_to_iter
from scrapy.utils.python import flatten
from w3lib.html import remove_tags

from .utils import identity, parse_json, replace_all_entities


class JsonLoader(ItemLoader):
    """ enhance ItemLoader with JMESPath capabilities """

    def __init__(self, response=None, json_obj=None, **kwargs):
        response = response if hasattr(response, "text") else None
        super(JsonLoader, self).__init__(response=response, **kwargs)

        if json_obj is None and response is not None:
            json_obj = parse_json(response.text)

        self.json_obj = json_obj
        self.context.setdefault("json", json_obj)

    def _get_jmes_values(self, jmes_paths):
        if self.json_obj is None:
            raise RuntimeError("no JSON object found")

        jmes_paths = arg_to_iter(jmes_paths)
        return flatten(
            jmespath.search(jmes_path, self.json_obj) for jmes_path in jmes_paths
        )

    def add_jmes(self, field_name, jmes, *processors, **kw):
        """ add values through JMESPath """

        values = self._get_jmes_values(jmes)
        self.add_value(field_name, values, *processors, **kw)

    def replace_jmes(self, field_name, jmes, *processors, **kw):
        """ replace values through JMESPath """

        values = self._get_jmes_values(jmes)
        self.replace_value(field_name, values, *processors, **kw)

    def get_jmes(self, jmes, *processors, **kw):
        """ get values through JMESPath """

        values = self._get_jmes_values(jmes)
        return self.get_value(values, *processors, **kw)


class GameLoader(ItemLoader):
    """ loader for GameItem """

    default_input_processor = MapCompose(
        identity, str, remove_tags, replace_all_entities, normalize_space
    )
    default_output_processor = TakeFirst()


class GameJsonLoader(JsonLoader, GameLoader):
    """ loader for GameItem plus JMESPath capabilities """


class UserLoader(ItemLoader):
    """ loader for UserItem """

    default_input_processor = MapCompose(
        identity, str, remove_tags, replace_all_entities, normalize_space
    )
    default_output_processor = TakeFirst()


class RatingLoader(ItemLoader):
    """ loader for RatingItem """

    default_input_processor = MapCompose(
        identity, str, remove_tags, replace_all_entities, normalize_space
    )
    default_output_processor = TakeFirst()


class RatingJsonLoader(JsonLoader, RatingLoader):
    """ loader for RatingItem plus JMESPath capabilities """



# -*- coding: utf-8 -*-

"""Logging classes."""

from scrapy.logformatter import LogFormatter


class QuietLogFormatter(LogFormatter):
    """Be quieter about scraped items."""

    def scraped(self, item, response, spider):
        return (
            super().scraped(item, response, spider)
            if spider.settings.getbool("LOG_SCRAPED_ITEMS")
            else None
        )
